from typing import List, Union
import sys
import os

current_dir = os.path.dirname(os.path.abspath(__file__))

from .classical_asp_encoding import enum_values, int_values, struct_values,\
    title_section, literal, to_asp_lines, action_causes, action_to_asp,\
    fluent_2_asp, param_val_2_asp, vars_to_asp, action_exec, equality_predicate

sys.path.insert(1, os.path.join(current_dir, '..', '..'))
from model import *


def operation_exec(action: Union[IAction, Method], exec_lit: Predicate) -> str:

    variables = set(action.params + exec_lit.variables)
    body = "" if len(variables) == 0 else ':-' + vars_to_asp(variables)
    s = ""
    if len(action.params) == 0:
        s = f'exec({action.name},{literal(exec_lit)})' + body
    else:
        parameters = ','.join(map(param_val_2_asp, action.params))
        s = f'exec({action.name}({parameters}),{literal(exec_lit)})' + body
    if isinstance(action, IAction):
        return 'action_' + s
    else:
        return 'method_' + s

def method_to_asp(action: Method, cond: str = None) -> str:
    name = action.name
    variables = action.params
    s = ''
    if len(action.params) == 0:
        s = f'method({name})'
        return s
    else:
        parameters = ','.join(map(param_val_2_asp, action.params))
        s = f'method({name}({parameters}))'
    if len(variables) > 0:
        if cond is None:
            return s + f':-{vars_to_asp(variables)}'
        else:
            return s + f':-{vars_to_asp(variables)}, {cond}'


def method_req(method: Method, prec: Goal, succ: Goal) -> str:
    name = method.name
    s = 'method_req('
    if len(method.params) == 0:
        s += f'{name}'
    else:
        parameters = ','.join(map(param_val_2_asp, method.params))
        s += f'{name}({parameters})'

    s += ', ' + goal(prec) + ', ' + goal(succ) + ')'

    if len(method.params) > 0:
        s += f':-{vars_to_asp(method.params)}'

    return s


def goal(g: Goal) -> str:
    name = g.name
    if len(g.arguments) == 0:
        return f'{name}'
    else:
        parameters = ','.join(map(param_val_2_asp, g.arguments))
        return f'{name}({parameters})'


def method_req_end(method: Method, maximal_goal: Goal) -> str:
    name = method.name
    s = 'method_req_end('
    if len(method.params) == 0:
        s += f'{name}'
    else:
        parameters = ','.join(map(param_val_2_asp, method.params))
        s += f'{name}({parameters})'

    s += ', ' + goal(maximal_goal) + ')'

    if len(method.params) > 0:
        s += f':-{vars_to_asp(method.params)}'

    return s


def goal_DNF(g: Goal, l: Literal) -> str:
    s = 'goal_DNF(' + goal(g) + ', ' + literal(l) + ')'
    variables = [v for v in g.arguments if isinstance(v, Variable)]
    if len(variables) > 0:
        s += f':-{vars_to_asp(variables)}'
    return s
    

def goal_def(g: Goal) -> str:
    variables = [v for v in g.arguments if isinstance(v, Variable)]
    s = 'goal(' + goal(g) + ')'
    if len(variables) > 0:
        s += f':-{vars_to_asp(variables)}'
    return s


def goal_to_sat(g: Goal) -> str:
    return 'goal_to_sat(' + goal(g) + ', 1, 1)'


def prec_to_sat(prec: Goal, succ: Goal) -> str:
    return 'prec_to_sat(' + goal(prec) + ', 1,' + goal(succ) + ', 1)'


def independent_rules() -> List[str]:

    return [
        "#program step(t)",
        "%creation of new subgoals",
        "goal_to_sat(G, t+1, t+1):-occurs(M, t), method_req(M, G, G2), goal(G)",
        "goal_to_sat(G, t+1, t+1):-occurs(M, t), method_req_end(M, G), goal(G)",
        "prec_to_sat(G1, t+1, G2, t+1):-occurs(M, t), method_req(M, G1, G2)",

        "%definition of minimal goal",
        "not_minimal(G2, T2, t):-prec_to_sat(G1, T1, G2, T2), goal_to_sat(G1,T1,t), goal_to_sat(G2, T2, t)",
        "not_minimal(G2, T2, t):-goal_to_sat(G1,T1,t), goal_to_sat(G2, T2, t), T2<T1",

        "minimal(G, T1, t):-goal_to_sat(G, T1, t), not not_minimal(G, T1, t)",

        "0{selected_sub_goal(G, T1, t):minimal(G, T1, t)}1",

        "%executability for methods:",
        "not_executable(M, t):-method_exec(M,F), fluent(F), not holds(F,t)",
        "not_executable(M, t):-method_exec(M,F), opposite(F, NOTF), holds(NOTF,t)",

        "not_relevant(M,t):-method(M), goal_DNF(MG,F), method_req_end(M, MG), selected_sub_goal(SG, ST, t), goal_DNF(SG, NOTF), opposite(F,NOTF)",
        "relevant(M,t):-method(M), selected_sub_goal(SG, ST, t), goal_DNF(SG, F), method_req_end(M, MG), goal_DNF(MG, F), not not_relevant(M,t)",
        "executable(M,t):-method(M), not not_executable(M,t), relevant(M,t)",
        ":- not executable(M,t), occurs(M,t), method(M)",

        "%executability for actions:",
        "not_executable(A,t):-exec(A,F), fluent(F), not holds(F,t)",
        "not_executable(A,t):-exec(A,NOTF), opposite(F, NOTF), holds(F,t)",

        "not_relevant(A,t):-action(A), selected_sub_goal(SG, ST, t), goal_DNF(SG, NOTF), causes(A,F), opposite(F, NOTF)",
        "relevant(A,t):-action(A), selected_sub_goal(SG, ST, t), goal_DNF(SG, F), causes(A,F), not not_relevant(A,t)",
        "executable(A, t):-action(A), not not_executable(A, t), relevant(A, t)",
        ":- not executable(A, t), occurs(A, t), action(A)",

        "%inertia for fluents:",
        "holds(F, t+1):-holds(F, t), opposite(F, G), not holds(G, t+1)",

        "%opposite definition",
        "opposite(F, neg(F)):-fluent(F)",
        "opposite(neg(F), F):-fluent(F)",

        "%at most one action/method at a time:",
        "{occurs(A,t):action(A)}",
        "{occurs(M,t):method(M)}",
        ":-occurs(A,t), occurs(M,t), A!=M",

        #no action nor method if selected goal is alredy satisfied:
        ":-occurs(O,t), now_already_sat(G, t), selected_sub_goal(G, T1, t)",

        "%action effects:",
        "holds(F, t+1):-action(A), occurs(A,t), causes(A,F)",

        "%inertia for subgoals",
        "action_occurs(t):-occurs(A,t), action(A)",

        "not_sat(G, t+1):-goal_DNF(G, F), holds(NOTF, t+1), opposite(NOTF ,F), goal(G)",
        "not_sat(G, t+1):-goal_DNF(G, F), fluent(F), not holds(F, t+1), goal(G)",

        "now_not_sat(G, t):-goal_DNF(G, F), holds(NOTF, t), opposite(NOTF ,F), goal(G)",
        "now_not_sat(G, t):-goal_DNF(G, F), fluent(F), not holds(F, t), goal(G)",

        "now_already_sat(G, t):- goal_to_sat(G, T1, t), not now_not_sat(G, t)",

        "not_to_propagate(G, T1, t):- now_already_sat(G, t), selected_sub_goal(G, T1, t)",

        "goal_to_sat(G, T1, t+1):- goal_to_sat(G, T1, t), not action_occurs(t), not not_to_propagate(G, T1, t)",
        "goal_to_sat(G, T1, t+1):- not_sat(G, t+1), goal_to_sat(G, T1, t), action_occurs(t)",
        "#program check(t)",
        ":- goal_to_sat(G, T, t), query(t)"
    ]


def compile_HGN_into_asp(problem: ClassicalPlanningProblem) -> str:


    s =  "%Answer set planning.\n\n"
    s += "%Answer set planning: A Survey. E. Pontelli et al. For a survey.\n"
    s += "%Epistemic Multiagent Reasoning with Collaborative Robots: D. Solda' el al. For a practical use-case.\n\n"

    s += title_section("[[\tPROBLEM DEPENDENT RULES\t]]")
    
    s += title_section('TYPES')
    type_asp_convertion = []
    for t in problem.types:
        if t.is_enum_type() or t.is_int_type() or t.is_struct_type():
            if t.is_enum_type():
                type_asp_convertion.append(f"{t.name}({enum_values(t)})")
            elif t.is_int_type():
                type_asp_convertion.append(f"{t.name}({int_values(t)})")
                

    s += to_asp_lines(type_asp_convertion)

    s += title_section('FLUENTS')

    fluent_asp = []
    for fluent in problem.fluents:
        fluent_asp.append(fluent_2_asp(fluent))

    s += to_asp_lines(fluent_asp)
    
    s += title_section('INITIALLY')

    init_values = []
    for lit in problem.init_values:
        init_values.append("holds(" + literal(lit) + ", 1)")

    s += to_asp_lines(init_values)

    s += title_section('ACTIONS')

    actions = []
    for action in problem.actions:
        action_equality_conds = []
        for precond in action.precondition:
            if isinstance(precond, EqualityPredicate):
                action_equality_conds.append(equality_predicate(precond))
        if len(action_equality_conds) == 0:
            actions.append(action_to_asp(action))
        else:
            actions.append(action_to_asp(action, ", ".join(action_equality_conds)))

    s += to_asp_lines(actions)

    s += title_section('EXECUTABLE')

    executabilities = []

    for action in problem.actions:
        for precond in action.precondition:
            if isinstance(precond, Literal):
                executabilities.append(action_exec(action, precond))

    s += to_asp_lines(executabilities)

    s += title_section('CAUSES')

    causes = []
    for action in problem.actions:
        for effect in action.effects:
            causes.append(action_causes(action, effect))

    s += to_asp_lines(causes)

    s += title_section('METHODS')

    methods = []

    for method in problem.methods:
        method_equality_conds = []
        print(method.name)
        for precond in method.precondition:
            if isinstance(precond, EqualityPredicate):
                method_equality_conds.append(equality_predicate(precond))
        if len(method_equality_conds) == 0:
            methods.append(method_to_asp(method))
        else:
            methods.append(method_to_asp(method, ", ".join(method_equality_conds)))

    #quit()
    s += to_asp_lines(methods)

    s += title_section('EXECUTABLE')

    executabilities = []

    for method in problem.methods:
        for precond in method.precondition:
            if isinstance(precond, Literal):
                executabilities.append(operation_exec(method, precond))
    s += to_asp_lines(executabilities)

    s += title_section('METHOD_GOAL_REFINEMENT')

    method_goal_refinement = []

    for method in problem.methods:
        for prec, succ in method.goal_poset.get_precedence_relations():
            method_goal_refinement.append(method_req(method, prec, succ))
        for maximal_goal in method.goal_poset.get_maximal_goals():
            method_goal_refinement.append(method_req_end(method, maximal_goal))
    s += to_asp_lines(method_goal_refinement)

    s += title_section('GOAL DNF')

    goal_dnf = []
    for g in problem.goals:
        for l in g.literals:
            goal_dnf.append(goal_DNF(g, l))

    s += to_asp_lines(goal_dnf)

    s += title_section('GOAL')

    goals = []
    for g in problem.goals:
        goals.append(goal_def(g))

    s += to_asp_lines(goals)

    s += title_section('GOAL TO SAT')

    init_poset = []
    initial_poset = problem.initial_poset
    for prec, succ in initial_poset.get_precedence_relations():
        init_poset.append(prec_to_sat(prec, succ))
    for goal in initial_poset.get_goals():
        init_poset.append(goal_to_sat(goal))

    s += to_asp_lines(init_poset)
    
    return s + to_asp_lines(independent_rules())



#!/usr/bin/env python
# Four spaces as indentation [no tabs]
import re

import itertools

class Action:

    #-----------------------------------------------
    # Initialize
    #-----------------------------------------------

    def __init__(self,  name, act_type, parameters, positive_preconditions, negative_preconditions, add_effects, del_effects, observers, p_observers, derive_cond = [], explicit_eff = [], extensions = None):
        def frozenset_of_tuples(data):
            return frozenset([tuple(t) for t in data])
        self.name = name
        self.act_type = act_type
        self.parameters = parameters
        self.positive_preconditions = positive_preconditions
        self.negative_preconditions = negative_preconditions
        self.add_effects = add_effects
        self.del_effects = del_effects
        self.observers = observers
        self.p_observers = p_observers
        self.derive_cond = derive_cond
        self.explicit_eff = explicit_eff

    #-----------------------------------------------
    # to String
    #-----------------------------------------------

    def __str__(self):
        return 'action: ' + self.name + \
        '\n  action type: ' + self.act_type + \
        '\n  parameters: ' + str(self.parameters) + \
        '\n  positive_preconditions: ' + str([list(i) for i in self.positive_preconditions]) + \
        '\n  negative_preconditions: ' + str([list(i) for i in self.negative_preconditions]) + \
        '\n  add_effects: ' + str([list(i[0]) + list(i[1]) for i in self.add_effects]) + \
        '\n  del_effects: ' + str([list(i[0]) + list(i[1]) for i in self.del_effects]) + \
        '\n  full observers: ' + str([list(i[0]) + list(i[1]) for i in self.observers]) + \
        '\n  partial observers: ' + str([list(i[0]) + list(i[1]) for i in self.p_observers]) + '\n'


    #-----------------------------------------------
    # Equality
    #-----------------------------------------------

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    #-----------------------------------------------
    # Groundify
    #-----------------------------------------------

    def groundify(self, objects, types, requirements, fluents):
        duplicates = True
        if ':no-duplicates' in requirements:
            duplicates = False
        if not self.parameters:
            yield self
            return
        type_map = []
        variables = []
        for var, type in self.parameters:
            type_stack = [type]
            items = []
            while type_stack:
                t = type_stack.pop()
                if t in objects:
                    items += objects[t]
                elif t in types:
                    type_stack += types[t]
                else:
                    raise Exception('Unrecognized type ' + t)
            type_map.append(items)
            variables.append(var)

        for assignment in itertools.product(*type_map):
            if (not duplicates and len(assignment) == len(set(assignment))) or duplicates:
                positive_preconditions = self.replace(self.positive_preconditions, variables, assignment, fluents, 1)
                negative_preconditions = self.replace(self.negative_preconditions, variables, assignment, fluents, 1)
                add_effects = self.pair_replace(self.add_effects, variables, assignment, fluents, 1)
                del_effects = self.pair_replace(self.del_effects, variables, assignment, fluents, 1)
                observers = self.pair_replace(self.observers, variables, assignment, fluents, 0)
                p_observers = self.pair_replace(self.p_observers, variables, assignment, fluents, 0)
                yield Action(self.name, self.act_type, assignment, positive_preconditions, negative_preconditions, add_effects, del_effects, observers, p_observers)

    #-----------------------------------------------
    # Replace
    #-----------------------------------------------

    def replace(self, group, variables, assignment,fluents, to_print):
        g = []
    #    print("Group: " + str(group))
        for pred in group:
            pred = list(pred)

            fluent = ''
            l=0
            j=''
            k=''

            #for i in pred:
            #    if 'C(' in i:
            #        j = re.sub(r'C\((.+)\,', r'C([\1],',i)
            #        pred[pred.index(i)] = j

            for i in pred:
                iv = 0
                l = l+1
                if 'B(' in i:
                    j = re.sub(r'B\((.+)\,', r'\1',i)
                    for t in variables:
                        if t == j:
                            pred[pred.index(i)] = 'B('+assignment[iv]+','
                    if type(pred[l]) is list:
                        tmp_list = self.replace([pred[l]],variables, assignment,fluents, to_print)
                        pred[l] = tmp_list[0]
                elif 'C(' in i:
                    k = re.sub(r'C\((.+)\,', r'\1',i)
                    for j in re.findall(r'[^,]+\,', k):
                        j = j.replace(',', '')
                        for t in variables:
                            if t == j:
                                tmp_str = pred[pred.index(i)]
                                tmp_str = tmp_str.replace(t,assignment[iv])
                                pred[pred.index(i)] = tmp_str
                    if type(pred[l]) is list:
                        tmp_list = self.replace([pred[l]],variables, assignment,fluents, to_print)
                        pred[l] = tmp_list[0]
                iv += 1

            for i in pred:
                iv = 0
                if 'C(' in i:
                    k = re.sub(r'C\((.+)\,', r'\1',i)
                    for j in re.findall(r'[^,]+\,', k):
                        j = j.replace(',', '')
                        for t in variables:
                            if t == j:
                                pred[pred.index(i)] = 'C(['+assignment[iv]+'],'
                iv += 1


            iv = 0
            for v in variables:
                while v in pred:
                    pred[pred.index(v)] = assignment[iv]
                iv += 1
            if to_print == 1:
                fluent = Action.unify_fluent_EFP(pred)
                if 'B(' not in fluent and 'C(' not in fluent and '-' not in fluent:
                    fluents.add(fluent)
            g.append(pred)
        return g

    def pair_replace(self, group, variables, assignment,fluents, to_print):
        g = []
        for elem in group:
            body = self.replace([elem[0]], variables, assignment, fluents, to_print)
            positive_head = self.replace(elem[1], variables, assignment, fluents, to_print)
            negative_head = self.replace(elem[2], variables, assignment, fluents, to_print)
            diff = self.replace(elem[3], variables, assignment, fluents, to_print)
            g.append((body[0],positive_head,negative_head,diff))
        return g

    @staticmethod
    def unify_fluent_EFP(given_list):
        fluent = ''
        l = 0
        parCount = 0
        while l < len(given_list):
            i = given_list[l]
            if 'C(' in i:
                i = i.replace('C(','C([')
                i = i[:-1]
                i = i + '],'

            if 'B(' in i or  'C(' in i:
                parCount +=1
                l += 1
                fluent += i

                if type(given_list[l]) is list:
                    fluent += Action.unify_fluent_EFP(given_list[l])
                    l += len(given_list[l])
                #fluent += (str(i))

            else:
                fluent += (str(i))
                if l != len(given_list) -1:
                    fluent += '_'
                l +=1

        while parCount != 0:
            fluent +=')'
            parCount -=1
        return fluent

    @staticmethod
    def unify_fluent_PDKB(given_list, no_change, from_bf):
        fluent = ''
        l = 0
        parCount = 0
        while l < len(given_list):
            i = given_list[l]
            if '-' in i and no_change == False:
                i = i.replace('-','!')

            if 'C(' in i:
                from_bf = True
                i = i.replace('C(','[')
                i = i[:-1]
                l += 1
                fluent += i + ']'

                if type(given_list[l]) is list:
                    fluent += Action.unify_fluent_PDKB(given_list[l], no_change,from_bf)
                    l += len(given_list[l])

            elif 'B(' in i:
                from_bf = True
                i = i.replace('B(','[')
                i = i[:-1]
                l += 1
                fluent += i + ']'

                if type(given_list[l]) is list:
                    fluent += Action.unify_fluent_PDKB(given_list[l], no_change,from_bf)
                    l += len(given_list[l])
                #fluent += (str(i))

            else:
                if from_bf:
                    fluent += '('
                    parCount += 1
                    from_bf= False
                fluent += (str(i))
                if l != len(given_list) -1:
                    fluent += ' '
                l +=1

        while parCount != 0:
            fluent +=')'
            parCount -=1

        return fluent

#-----------------------------------------------
# Main
#-----------------------------------------------
if __name__ == '__main__':
    a = Action('move', [['?ag', 'agent'], ['?from', 'pos'], ['?to', 'pos']],
        [['at', '?ag', '?from'], ['adjacent', '?from', '?to']],
        [['at', '?ag', '?to']],
        [['at', '?ag', '?to']],
        [['at', '?ag', '?from']]
    )
    print(a)

    objects = {
        'agent': ['ana','bob'],
        'pos': ['p1','p2']
    }
    types = {'object': ['agent', 'pos']}
    for act in a.groundify(objects, types):
        print(act)



"""
module for the definiton of Types of parameters in the planning problem
***This way to proceed means that we :require :typing
"""
from typing import List, Tuple
from dataclasses import dataclass
from abc import abstractmethod


@dataclass(frozen=True)
class Type:
    """Definition of the Type class"""
    name: str

    def is_bool_type(self) -> bool:
        """is boolean?"""
        return False

    def is_enum_type(self) -> bool:
        """is enum?"""
        return False

    def is_int_type(self) -> bool:
        """is int?"""
        return False

    def is_struct_type(self) -> bool:
        """is struct ?"""
        return False

    def is_agent_type(self) -> bool:
        """we reserve a particular type for agents"""
        return False

    def __eq__(self, _type):
        return self is _type or \
            _type.is_bool_type() and self.is_bool_type()

    def __len__(self):
        return 1

    @abstractmethod
    def __hash__(self) -> int:
        pass


@dataclass(frozen=True)
class BoolType(Type):
    """Definition of boolean type"""
    def is_bool_type(self) -> bool:
        return True

    def __repr__(self) -> str:
        return f"{self.name} bool"

    def __hash__(self) -> int:
        return 100

    def __len__(self):
        return 0

@dataclass(frozen=True)
class IntType(Type):
    """Definition of integer type"""
    name: str
    min_val: int
    max_val: int

    def __post_init__(self):

        if self.max < self.min:
            raise Exception(f"max : {self.max} > min : {self.min}")

        if not isinstance(self.min, int) or not isinstance(self.max, int):
            raise Exception(f"integer type ill-defined: IntType({self.min}, {self.max})")

    @property
    def min(self) -> int:
        """Returns the minimum range value"""
        return self.min_val

    @property
    def max(self) -> int:
        """Returns the maximum range value"""
        return self.max_val

    def __contains__(self, i: int):
        return self.min <= i <= self.max

    def __repr__(self) -> str:
        return f"{self.name} : [{self.min}, {self.max}]"

    def is_int_type(self) -> bool:
        return True

    def contained_in(self, other: "IntType") -> bool:
        """Returns if the range is contained"""
        return self.min >= other.min and self.max <= other.max

    @property
    def interval(self) -> Tuple[int]:
        """Returns the interval"""
        return (self.min, self.max)

    def __hash__(self) -> int:
        h_name = hash(self.name)
        return 31**self.min**self.max+h_name % 256


@dataclass(frozen=True)
class EnumType(Type):
    """Definition of enumerate type"""
    name: str
    domain: List[str]

    def __post_init__(self):

        if not isinstance(self.domain, list):
            raise Exception(f"{domain} is not a list of strings")
        for enum_val in self.domain:
            if not isinstance(enum_val, str):
                raise Exception(f"{enum_val} is not a string")

    def __repr__(self) -> str:
        return f"{self.name} : {self.domain}"

    def is_enum_type(self) -> bool:
        return True

    def __iter__(self):
        for val in self.domain:
            yield val

    def contained_in(self, other: "EnumType") -> bool:
        """Returns if the domain is contained"""
        for s in self.domain:
            if s not in other.domain:
                return False
        return True

    @property
    def enum_values(self) -> List[str]:
        """Return the domain"""
        return self.domain

    def __hash__(self) -> int:
        return 1
        h = 1
        prime = 31
        for val in self.domain[1:]:
            h = prime*h+hash(val)
        return hash(self.name) % 256 + h


@dataclass(frozen=True)
class AgentType(EnumType):
    """A type suited for agents"""
    name: str
    domain: List[str]

    def __init__(self, domain):
        super().__init__('agent', domain)

    def __post_init__(self):
        super().__post_init__()

    def is_agent_type(self):
        return True

    def __repr__(self):
        return super().__repr__()


@dataclass(frozen=True)
class StructType(Type):
    """Definition of struct type"""
    name: str
    subtypes: List[Type]

    def __post_init__(self):

        if not isinstance(self.subtypes, list):
            raise Exception(f"""{self.subtypes} is not a list of plan ftype.Type
 {{ftype.EnumType, ftype.BoolType, ftype.IntType}}""")

        for sub_type in self.subtypes:
            if not isinstance(sub_type, Type):
                raise Exception(f"{sub_type} is not an instance of ftype.Type")
            if isinstance(sub_type, StructType):
                raise Exception(f"nested StructType are not supported")
            if isinstance(sub_type, BoolType):
                raise Exception(f"""ftype.BoolType cannot be a field in a
fType.StructType""")

    def __repr__(self):
        return f"{self.name} : {self.subtypes}"

    def is_struct_type(self) -> bool:
        return True

    def __iter__(self):
        for subtype in self.subtypes:
            yield subtype

    def __len__(self):
        return len(self.subtypes)

    def __hash__(self) -> int:#TODO
        h = 1
        return h



from setuptools import setup, find_packages

with open("ECHO/README.org", "r") as ldf:
    long_description = ldf.read()

print(find_packages())

setup(
    name='echosys',
    version='0.1.21',
    description='A python library for modeling and solving epistemic multi agents planning problems',
    long_description = long_description,
    long_description_content_type = '',
    url='https://github.com/DavideSolda/echosys',
    author='Davide Solda\'',
    author_email='davide.solda@tuwien.ac.at',
    license='GNU General Public License v3.0',
    packages=find_packages(),
    install_requires=['clingo'],
    classifiers=[]
    )



import sys, os
from dataclasses import field, dataclass
from typing import List, Union, ClassVar, Optional, Dict
import copy
from enum import Enum

current_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(1, current_dir)

from .arithmetic_expression import is_int_fvalue, is_enum_fvalue, ArithmeticExpr
from .variable import Variable
from .ftype import Type, AgentType


@dataclass
class Predicate():


    negated: bool
    variables: List[Variable]
    types: List[Union[int, str, Variable]]

    def __post_init__(self):
        self.negated = False

    def __neg__(self) -> 'Predicate':
        negated_p = copy.deepcopy(self)
        negated_p.negated = not self.negated
        return negated_p

    def __and__(self, pl: 'Predicate', pr: 'Predicate') -> 'BooleanPredicate':
        return BooleanPredicate(BooleanOperator.AND, pl, pr)

    def __or__(self, pl: 'Predicate', pr: 'Predicate') -> 'BooleanPredicate':
        print(BooleanPredicate(BooleanOperator.OR, pl, pr))
        quit()
        return BooleanPredicate(BooleanOperator.OR, pl, pr)




class BooleanOperator(Enum):

    AND = 'and'
    OR = 'or'


@dataclass
class BooleanPredicate(Predicate):

    op: BooleanOperator
    left_predicate: Predicate
    right_predicate: Predicate
    
    def __post_init__(self):
        """just add info about variables and types"""
        super().__post_init__()
        self.variables = self.left_predicate.variables + self.right_predicate.variables
        self.types = self.left_predicate.types + self.right_predicate.types


def conj(lp, rp) -> BooleanPredicate:
    return BooleanPredicate(False, [], [], BooleanOperator.AND, lp, rp)


def disj(lp, rp) -> BooleanPredicate:
    return BooleanPredicate(False, [], [], BooleanOperator.OR, lp, rp)


@dataclass
class Literal(Predicate):

    fluent: 'Fluent'
    args: List[Union[ArithmeticExpr, int, str, Variable]] = field(
        default_factory=(lambda: []))

    def __post_init__(self):
        """just add info about variables and types"""
        super().__post_init__()

        self.types = [self.fluent.type]
        self.variables = [arg.variables for arg in self.args
                          if isinstance(arg, ArithmeticExpr)]
        self.variables += [arg for arg in self.args
                           if isinstance(arg, Variable)]

    def instatiate(self, var_val: Dict[Variable, Union[str, int]]) -> 'Literal':

        args = copy.deepcopy(self.args)
        for idx, arg in enumerate(args):
            if isinstance(arg, Variable):
                args[idx] = var_val[arg]
        if self.negated:
            return -self.fluent(*args)
        else:
            return self.fluent(*args)

class EqualityOperator(Enum):

    eq = "=="
    neq = "!="


@dataclass
class EqualityPredicate(Predicate):


    operator: EqualityOperator
    left_operand: Union[ArithmeticExpr, int, Variable]
    right_operand:Union[ArithmeticExpr, int, Variable]

    def __post_init__(self):

        super().__post_init__()
        
        self.variables = get_fvalues_variable(self.left_operand) +\
            get_fvalues_variable(self.right_operand)
        self.types = get_fvalues_type(self.left_operand) +\
            get_fvalues_type(self.right_operand)
        if is_int_fvalue(self.left_operand) and is_int_fvalue(self.right_operand) or\
           is_enum_fvalue(self.left_operand) and is_enum_fvalue(self.right_operand):
            return
        else:
            raise Exception(f"{l} and {r} are not compatible")

    def __repr__(self) -> str:
        return f'sc.EqualityPredicate: {self.left_operand}' + str(self.operator) + f'{self.right_operand}'


def eq(l: Union[ArithmeticExpr, int, Variable, str],
       r: Union[ArithmeticExpr, int, Variable, str]):

    return EqualityPredicate(negated = False, types = [], variables = [],
                             operator = EqualityOperator.eq, left_operand = l,
                             right_operand = r)


def neq(l: Union[ArithmeticExpr, int, Variable, str],
        r: Union[ArithmeticExpr, int, Variable, str]):

    return EqualityPredicate(negated = False, types = [], variables = [],
                             operator = EqualityOperator.neq, left_operand = l,
                             right_operand = r)


def get_fvalues_type(fvalue: Union[ArithmeticExpr, int, Variable, str]) -> List[Type]:
    if isinstance(fvalue, ArithmeticExpr):
        return fvalue.types
    elif isinstance(fvalue, Variable):
        return [fvalue.type]
    return []


def get_fvalues_variable(fvalue: Union[ArithmeticExpr, int, Variable, str]) -> List[Variable]:
    if isinstance(fvalue, Variable):
        return [fvalue]
    if isinstance(fvalue, ArithmeticExpr):
        return fvalue.variables
    return []


class BeliefPredicate(Predicate):
    """B_ag(current_position(4))"""
    agents: List[Union[Variable, str]]
    belief_proposition: Predicate

    def __init__(self, agents: List[Union[Variable, str]],
                 proposition: Predicate):

        for agent in agents:
            assert isinstance(agent, (Variable, str))
        assert isinstance(proposition, Predicate)

        types = proposition.types
        
        #  deal with variables
        variables = proposition.variables
        for agent in agents:
            if isinstance(agent, Variable):
                variables.append(agent)

        super().__init__(negated=False, variables=variables, types=types)
        #  typing checks
        for agent in agents:
            assert isinstance(agent, Variable) and agent.is_agent() or \
                isinstance(agent, str)

        assert isinstance(proposition, Literal) \
            or type(proposition) == type(self)

        self.agents = agents
        self.belief_proposition = proposition


    def __repr__(self) -> str:
        s = f'B_{self.agents} ({self.belief_proposition})'
        return f'not {s}' if self.negated else s


def B(agents, proposition) -> BeliefPredicate:
    if type(proposition).__name__ == 'Fluent':
        proposition = Literal(fluent=proposition, args=[])
    return BeliefPredicate(agents=agents, proposition=proposition)


@dataclass
class When(Predicate):
    body: Predicate
    head: Predicate

    def __init__(self, body, head):

        negate = False
        variables = body.variables + head.variables
        types = body.types + head.types

        super().__init__(negate, variables, types)

        assert isinstance(body, Predicate)
        assert isinstance(head, Predicate)

        self.variables = body.variables + head.variables
        self.types = body.types + head.types

        self.body = body
        self.head = head

class Forall():
    """Predicate to express forall formulae"""

    quantified_variable: Variable
    disequality_predicate: Optional[EqualityPredicate]
    when: Optional[When]
    who: Union[Variable, str]

    variables: List[Variable]
    types: List[Type]

    def __init__(self, quantified_variable: Variable, neq: EqualityPredicate=None,
                 when: Predicate=None, who: Union[Variable, str]=None):

        self.quantified_variable = quantified_variable
        self.disequality_predicate = neq
        self.when = when
        self.who = who


    def __post_init__(self):

        self.variables = []
        self.types = []

        #check that the quantified variable is an agent
        if not isinstance(self.quantified_variable, Variable):
            raise ValueError(f'''{quantified_variable} is neither the name of an agent,
            nor an agent variable''')
        else:
            self.variables.append(self.quantified_variable)

        #if specified, equality_predicate should contain the quantified variable
        if self.equality_predicate is not None:
            assert isinstance(self.equality_predicate, EqualityPredicate)

            if disequality_predicate.operator == EqualityOperator.eq or \
               quantified_variable not in disequality_predicate.variables:
                raise ValueError(f'''forall {disequality_predicate.variables[0].name} == 
                {disequality_predicate.variables[1].name} {quantified_variable.name} does not make sense)''')
            self.variables = disequality_predicate.variables

        #if specified, when should be an instance of Predicate
        if when is not None and not isinstance(when, Predicate):
            raise ValueError(f'when should be a predicate: {when}')
        elif when is not None:
            self.variables += [var for var in when.variables
                               if var != self.quantified_variable]


    def __repr__(self) -> str:
        
        s = f'forall {self.quantified_variable.name} '
        if self.disequality_predicate is not None:
            s += f'{self.disequality_predicate} '
        if self.when is not None:
            s += f'(when {self.when} )'
        s += f'{self.quantified_variable.name}' 
        return s
